#include <gtk/gtk.h>
#include <gtkmozembed.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "light.h"

#include "fifo.h"
#include "moz_cb.h"
#include "moz_main.h"
#include "moz_pref.h"
#include "pref.h"
#include "ui.h"
#include "utils.h"

static void check_running(char *path, const char *newurl);

/* global variables */
int win_count = 0;
GList *win_list = NULL;
Config config;

/* character set detection stuff */
GHashTable *detect = NULL;
#ifndef ENGLISH_IS_NO1
char **charsets = NULL;
#endif

/* edit as necessary */
static char *searchpath[] = {
    "/usr/local/mozilla",		/* me */
    "/usr/lib/mozilla",			/* monkey-spank */
    "/opt/mozilla",			/* roothat */
    "/usr/lib/mozilla-0.9.1",		/* crapian */
    NULL
};

/* this assumes mozilla libs are in ld.so.conf */
static void light_wrapper(void)
{
    int i = 0;
    if (getenv("MOZILLA_FIVE_HOME") == NULL) {
	while (searchpath[i]) {
	    if (!access(g_strconcat(searchpath[i], "/chrome/embed.jar", NULL),
			F_OK | R_OK)) {
		/* found it */
		putenv(g_strconcat("MOZILLA_FIVE_HOME=", searchpath[i], NULL));
		return;
	    }
	    i++;
	}
	g_print("Cannot find Mozilla installation directory.  Please set\n"
		"MOZILLA_FIVE_HOME to your Mozilla directory.\n");
	exit(1);
    }
}

static void check_running(char *path, const char *newurl)
{
    int pid;
    int oldpid;
    FILE *fp;

    pid = getpid();
    fp = fopen(path, "r");
    
    /* previous pid file found */
    if (fp) {
	fscanf(fp, "%d", &oldpid);
	fclose(fp);
	if (kill(oldpid, 0)) {
	    /* stale pid */
	    unlink(path);
	} else {		/* it's still running */
	    if (newurl) {
		/* got a URL to load in the new window */
		char *urlpath;
		FILE *urlfp;
		/* ~/.light/lock_url */
		urlpath = g_strconcat(path, "_url", NULL);
		urlfp = fopen(urlpath, "w");
		if (urlfp) {
		    fprintf(urlfp, "%s\n", newurl);
		    fclose(urlfp);
		}
		g_free(urlpath);
	    }
	    /* get a new window from the previously running process
	     * and exit this one */
	    kill(oldpid, SIGUSR1);
	    g_free(path);
	    exit(0);
	}
    }
    /* pid file wasnt found, so make a new one */
    fp = fopen(path, "w");
    fprintf(fp, "%d", pid);
    fclose(fp);
    g_free(path);
}

int main(int argc, char **argv)
{
    char *home = NULL;			/* path to homedir */
    char *pref = NULL;			/* path to prefs file */
    int first = 0;			/* first run, no prefs? */
    int c = 0;
    Browser *browser = NULL;

    gtk_set_locale();
    gtk_init(&argc, &argv);

    memset(&config, 0, sizeof(Config));

    home = g_get_home_dir();
    if (!home) {
	g_print("Failed to get HOME\n");
	exit(1);
    }

    /* process any leftover command line switches */
    while ((c = getopt(argc, argv, "v")) != EOF) {
	switch (c) {
	    case 'v':
		g_print("Light version " LIGHT_VERSION "\n");
		exit(0);
		break;
	}
    }
    argc -= optind;
    argv += optind;

    config.pref = g_strconcat(home, "/.light", NULL);
    mkdir(config.pref, 0755);
    g_free(home);

    light_wrapper();

    /* see if Light is already running elsewhere and popup a new window */
    check_running(g_strconcat(config.pref, "/lock", NULL),
	    (argc > 0) ? argv[0] : NULL);

    win_list = g_list_alloc();

    pref = g_strconcat(config.pref, "/prefs.js", NULL);
    first = access(pref, F_OK | R_OK);
    g_free(pref);

    /* initialize embedding API */
    gtk_moz_embed_set_profile_path(config.pref, "");
    gtk_moz_embed_push_startup();
    embed_single_init();
    /* register some custom handlers */
    moz_init(NULL);
    light_set_preferences(first);
#ifdef DOWNLOAD_HANDLER
    /* initialize "smart" download handler */
    moz_mime_init();
#endif /* DOWNLOAD_HANDLER */

    /* load up fifo entries from last session */
    fifo_load();
    /* make up detect menu hash table */
    light_make_detect();

    browser = light_new(GTK_MOZ_EMBED_FLAG_DEFAULTCHROME);
    gtk_widget_set_usize(browser->embed, 500, 500);
    light_set_visibility(browser, TRUE);
    moz_history_init(browser);

    if (argc > 0)
	gtk_moz_embed_load_url(GTK_MOZ_EMBED(browser->embed), argv[0]);
    else {
	char *homepage;
	moz_pref_get_char(LT_HOMEPAGE, &homepage);
	gtk_moz_embed_load_url(GTK_MOZ_EMBED(browser->embed), homepage);
	g_free(homepage);
    }

    /* catch new window handler */
    signal(SIGUSR1, sigusr1_handler);

    gtk_main();

    /* quit, cleanup */
    moz_pref_save();
    fifo_halt();
    g_hash_table_destroy(detect);
    g_free(config.pref);
    gtk_moz_embed_pop_startup();

    return 0;
}
