#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gtkmozembed.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>
#include <zlib.h>
#include "light.h"

#include "html.h"
#include "ui.h"
#include "utils.h"

extern Config config;
extern GHashTable *detect;

static const MenuEntry autodetect[] = {
    { "", "<main>/Page/Autodetect/Off" },
    { "zh_parallel_state_machine", "<main>/Page/Autodetect/Chinese" },
    { "cjk_parallel_state_machine", "<main>/Page/Autodetect/East Asian" },
    { "ja_parallel_state_machine", "<main>/Page/Autodetect/Japanese" },
    { "ko_parallel_state_machine", "<main>/Page/Autodetect/Korean" },
    { "ruprob", "<main>/Page/Autodetect/Russian" },
    { "zhcn_parallel_state_machine", "<main>/Page/Autodetect/Simplified Chinese" },
    { "zhtw_parallel_state_machine", "<main>/Page/Autodetect/Traditional Chinese" },
    { "ukprob", "<main>/Page/Autodetect/Ukrainian" },
    { NULL, NULL }
};

static void light_selection_get(GtkWidget *widget, GtkSelectionData *data,
	unsigned int info, unsigned int time_stamp, gpointer noise);
static int string_cmp (gconstpointer a, gconstpointer b);

typedef enum {
    SEL_TYPE_NONE,
    APPLE_PICT,
    ATOM,
    ATOM_PAIR,
    BITMAP,
    C_STRING,
    COLORMAP,
    COMPOUND_TEXT,
    DRAWABLE,
    INTEGER,
    PIXEL,
    PIXMAP,
    SPAN,
    STRING,
    TEXT,
    WINDOW,
    LAST_SEL_TYPE
} SelType;

static GtkTargetEntry targetlist[] = {
    { "STRING", 0, STRING },
    { "COMPOUND_TEXT", 0, COMPOUND_TEXT }
};
static int ntargets = sizeof(targetlist) / sizeof(targetlist[0]);

static void light_selection_get(GtkWidget *widget, GtkSelectionData *data,
	unsigned int info, unsigned int time_stamp, gpointer noise)
{
    char *text;

    text = gtk_object_get_data(GTK_OBJECT(widget), "selection");

    g_return_if_fail(text != NULL);

    gtk_selection_data_set(data, GDK_SELECTION_TYPE_STRING,
			   8, text, strlen(text));
}

void light_selection_setup(Browser *browser)
{
    /* setup selection targets */
    gtk_selection_add_targets(browser->window, GDK_SELECTION_PRIMARY,
			      targetlist, ntargets);
    gtk_signal_connect(GTK_OBJECT(browser->window), "selection_get",
		       GTK_SIGNAL_FUNC(light_selection_get), browser);
}

void light_copy_clipboard(Browser *browser, const char *text)
{
    if (browser->clipboard)
	g_free(browser->clipboard);

    browser->clipboard = g_strdup(text);

    gtk_object_set_data(GTK_OBJECT(browser->window), "selection",
			browser->clipboard);

    gtk_selection_owner_set(browser->window,
			    GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME);
}

void update_message(Browser * browser, const char *message)
{
    if (browser->sb.link)
	g_free(browser->sb.link);
    
    if (message)
	browser->sb.link = g_strdup(message);
    else
	browser->sb.link = NULL;

    light_update_status(browser);
}

char *uncompress_stream(int streamid)
{
    const char *streams[] = { about_stream, manual_stream };
    const int csize[] = { ABOUT_CLEN, MANUAL_CLEN };
    const int size[] = { ABOUT_LEN, MANUAL_LEN };

    char *dest;
    long int len;

    dest = g_new0(char, size[streamid]);
    len = size[streamid];

    uncompress(dest, &len, streams[streamid], csize[streamid]);
    return dest;
}

static int string_cmp (gconstpointer a, gconstpointer b)
{
    return (g_strcasecmp(a, b) == 0);
}

void light_make_detect(void)
{
    int i;
    detect = g_hash_table_new(g_str_hash, string_cmp);
    g_hash_table_freeze(detect);
    for (i = 0; autodetect[i].name != NULL; i++) {
	g_hash_table_insert(detect, autodetect[i].path,
		autodetect[i].name);
    }
    g_hash_table_thaw(detect);
}

/**
 * set a mark in the autodetect menu... This is pretty stupid, since
 * hash table only allows to lookup value by key but not the other way
 * around :(
 */
void light_set_detect(gpointer k, gpointer v, gpointer d)
{
    char *key = (char *)k;
    char *value = (char *)v;
    Browser *browser = (Browser *)d;
    GtkWidget *menu;

    if (!g_strcasecmp(config.detector, value)) {
	/* skip <main> */
	key += 6;
	menu = gtk_item_factory_get_widget(browser->fact, key);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menu), TRUE);
    }
}

/**
 * url_basename: parse url like http://some.com/any/here?query=lameness+stuff
 * and return "here"
 */
char *url_basename(const char *input)
{
    int run;
    char *top;
    char *string;
    char *result;

    if (!input)
	return NULL;

    /* make a copy to work on */
    string = g_strdup(input);
    top = string;

    /* run backwards and look for beginning of CGI params */
    run = strlen(string) + 1;
    while (--run > 0) {
	if (*string == '?') {
	    *string = '\0';
	    break;
	}
	string++;
    }
    result = g_strdup(g_basename(top));
    g_free(top);
    return result;
}

/**
 * escape_string: escape a string replacing non-ascii with %xx
 * thanks, cURL. Returned string should be freed.
 */
char *escape_string(const char *string, int length)
{
    int alloc = (length ? length : strlen(string)) + 1;
    char *ns = g_new(char, alloc);
    unsigned char in;
    int newlen = alloc;
    int index = 0;

    length = alloc - 1;
    while (length--) {
	in = *string;
	if(' ' == in)
	    ns[index++] = '+';
	else if (!(in >= 'a' && in <= 'z') &&
		!(in >= 'A' && in <= 'Z') &&
		!(in >= '0' && in <= '9')) {
	    /* encode it */
	    if (('%' == in) && (length >= 2) &&
		    isxdigit((int)string[1]) &&
		    isxdigit((int)string[2]) ) {
		/* This is an already encoded letter, leave it! */
		memcpy(&ns[index], string, 3);
		string += 2;
	    } else {
		/* encode this now */
		newlen += 2; /* the size grows with two, since this'll become a %XX */
		if (newlen > alloc) {
		    alloc *= 2;
		    ns = g_renew(char, ns, alloc);
		    if (!ns)
			return NULL;
		}
		g_snprintf(&ns[index], 4, "%%%02X", in);
	    }
	    index += 3;
	} else {
	    /* just copy this */
	     ns[index++]=in;
	}
	string++;
    }
    ns[index] = 0; /* terminate it */
    return ns;
}

/**
 * light_download: start external downloader with the link in @url.
 */
void light_download(Browser *browser, const char *uri)
{
    char *cmdline;

    /* download begins */
    if (strstr(config.handler, "%s")) {
	cmdline = g_strdup_printf(config.handler, uri);
	/* run it ... lame, but effective */
	system(cmdline);
	g_free(cmdline);
    } else {
	/* when called from content_handler we don't have our browser
	 * pointer. So, do nothing */
	if (browser) {
	    update_message(browser,
		"Download handler not configured - RTFM");
	}
    }
}

/* cute signal handler to open up a new window */
void sigusr1_handler(int sig)
{
    FILE *fp;
    char lockurl[512];
    char *file;

    memset(lockurl, 0, sizeof(lockurl));
    file = g_strconcat(config.pref, "/lock_url", NULL);
    fp = fopen(file, "r");
    if (fp) {
	fgets(lockurl, 511, fp);
	g_strchomp(lockurl);
	if (lockurl[0])
	    light_window_new(lockurl);
	fclose(fp);
    } else {
	light_window_new("about:blank");
    }
    unlink(file);
    g_free(file);
}

void dialog_esc_handler(GtkWidget *widget, GdkEventKey *key, GtkWidget *window)
{
    /* GTK really sucks.  This is what we have to do to get normal
     * user accepted behavior of pressing "esc" key to get rid of a dialog
     * box */
    if (key->keyval == GDK_Escape)
	gtk_widget_destroy(window);
}

char *check_file_exists(const char *directory, const char *filename)
{
    char *full_filename;
    struct stat s;
    int status;

    full_filename = g_new0(char, strlen(directory) + 1 + strlen(filename) + 1);
    strcpy(full_filename, directory);
    strcat(full_filename, G_DIR_SEPARATOR_S);
    strcat(full_filename, filename);

    status = stat(full_filename, &s);
    if (status == 0 && S_ISREG(s.st_mode))
	return full_filename;
    g_free(full_filename);
    
    return NULL;
}
